#!# Common code used by snapshot system clients
#!#
#!# Copyright (C) 2002 Kyle E. Jones
#!#
#!# This program is free software; you can redistribute it and/or modify
#!# it under the terms of the GNU General Public License as published by
#!# the Free Software Foundation; either version 1, or (at your option)
#!# any later version.
#!#
#!# This program is distributed in the hope that it will be useful,
#!# but WITHOUT ANY WARRANTY; without even the implied warranty of
#!# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#!# GNU General Public License for more details.
#!#
#!# You should have received a copy of the GNU General Public License
#!# along with this program, in a file named COPYING.

use Getopt::Long;

sub
diag
{
   print STDERR "$progname: ", @_, "\n";
}

sub
fatal
{
   diag @_;
   delete_files;
   exit 1;
}

my ($user,undef,undef,undef,undef,undef,undef,$home,undef) = getpwuid($<);
fatal "getpwuid failed: $!" if $user eq '';
my $registration_file = $home . '/' . '.snap_registrations';

my $exclusion_file = $home . '/' . '.snap_exclusions';
$exclusion_file = undef if ! -f $exclusion_file;

my $debug = 0;

autoflush STDERR 1;

sub
send_command
{
   my $first = shift @_;
   my $str = join(' ', $first, map { safety_encode($_) } @_) . "\r\n";
   print TOBACKEND $str;
   return if ! $debug;
   $str =~ s/\r\n/\n/;
   print STDERR "--> $str";
}

sub
read_response
{
   # only return the final OK or NO response unless first arg is
   # non-zero, in which case return every line.
   my ($return_all) = @_;
   my $line;
   do {
      $line = <FROMBACKEND>;
      $line =~ s/[\r\n]//g;
      print STDERR "<-- $line\n" if $debug;
   } while (defined $line && (! $return_all && $line !~ /^(OK|NO)( .?)?/));
   return $line;
}

sub
read_registrations_xxx
{
   my ($lock) = (@_);

   if ($lock) {
      # first create the file is it does not exist
      open REG, ">>$registration_file";
      close REG;
      # now open the file read/write
      open REG, "+<$registration_file" or
         fatal "couldn't open registration file $registration_file (rw): $!";
      flock REG, $LOCK_EX;
   } else {
      open REG, "<$registration_file" or
         fatal "couldn't open registration file $registration_file (r): $!";
      flock REG, $LOCK_SH;
   }
   my $line;
   my $r;
   my $lineno = 0;
   while (defined($line = <REG>)) {
      chomp $line;
      $lineno++;
      next if $line =~ /^[ \t]*$/;
      next if $line =~ /^#/;
      my ($source, $id, $user, $host, $repositories) =
         map { safety_decode($_) } split /\037/, $line;
      if ($source eq '' || $id eq '' || $user eq '' || $repositories eq '') {
         fatal "$registration_file: line $lineno: too few fields";
      }
      $r->{$source}{source} = $source;
      $r->{$source}{id} = $id;
      $r->{$source}{user} = $user;
      $r->{$source}{remotehost} = $host;
      $r->{$source}{repositories} = $repositories;
   }

   # leave REG open to keep the lock in place
   if (! $lock) {
      close REG;
   }

   return $r;
}

sub
read_registrations
{
   return read_registrations_xxx(0);
}

sub
read_and_lock_registrations
{
   return read_registrations_xxx(1);
}

sub
write_registrations
{
   my ($r) = (@_);
   my ($key);

   truncate REG, 0;
   seek REG, 0, 0;
   print REG "# IF YOU DELETE THIS FILE THE SNAPSHOT SYSTEM WILL LOSE\n";
   print REG "# TRACK OF ALL YOUR BACKUPS.  DON'T DELETE THIS FILE\n";
   print REG "# UNLESS YOU'RE SURE OF WHAT YOU'RE DOING.\n";
   foreach $key (keys %{ $r }) {
      print REG $r->{$key}{source}, "\037";
      print REG $r->{$key}{id}, "\037";
      print REG $r->{$key}{user}, "\037";
      print REG $r->{$key}{remotehost}, "\037";
      print REG $r->{$key}{repositories}, "\n";
   }
   # closing the file unlocks it.
   if (! close REG) {
      fatal "write to registration file failed: $!";
   }
}

# is $b a subdirectory of $a
sub
subdirof
{
   my ($a, $b) = @_;

   while ($b ne '') {
      if ($a eq $b) {
         return 1;
      }
      $b =~ s,/[^/]*$,,;
   }
   return 0;
}

sub
hostmatch
{
   my ($a, $b, $domain) = @_;

   $a .= $domain if $a !~ /\./;
   $b .= $domain if $b !~ /\./;
   $a = lc($a);
   $b = lc($b);
   return $a eq $b;
}
