#!# Common code used by the snapshot system
#!#
#!# Copyright (C) 2002 Kyle E. Jones
#!#
#!# This program is free software; you can redistribute it and/or modify
#!# it under the terms of the GNU General Public License as published by
#!# the Free Software Foundation; either version 1, or (at your option)
#!# any later version.
#!#
#!# This program is distributed in the hope that it will be useful,
#!# but WITHOUT ANY WARRANTY; without even the implied warranty of
#!# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#!# GNU General Public License for more details.
#!#
#!# You should have received a copy of the GNU General Public License
#!# along with this program, in a file named COPYING.

use strict;
use Cwd;
use IPC::Open2;
use FileHandle;

my $progname = $0;
$progname =~ s,.*/,,;

my $source_info_file = "source";
my $encoder_info_file = "encoder";
my $diskusage_info_file = "diskusage";

my $snap_backend_program = 'snap-backend';

my $ssh_program = 'ssh';
my $rsync_program = 'rsync';
my $find_program = 'find';
my $rm_program = 'rm';
my $hostname_program = 'hostname';

my $mount_program;
{
   my $x;
   foreach $x (qw(/sbin/mount /usr/sbin/mount /usr/etc/mount)) {
      if (-x $x) {
         $mount_program = $x;
         last;
      }
   }
}
$mount_program = 'mount' if ! defined $mount_program;

# for use with flock
my $LOCK_SH = 1;
my $LOCK_EX = 2;
my $LOCK_NB = 4;
my $LOCK_UN = 8;

srand (time() ^ ($$ + (getppid() << 17) + ($$ << 15)));

sub
safety_encode
{
  my ($str) = @_;
  $str =~ s/%/%25/g;
  $str =~ s/[\000-\040\177]/sprintf "%%%02x", ord($&)/eg;
  return $str;
}

sub
safety_decode
{
  my ($str) = @_;
  $str =~ s/%([0-9A-F][0-9A-F])/chr(hex($1))/eg;
  return $str;
}

sub
quincy
{
   my ($code) = @_;

   $code = $? if ! defined $code;
   $code &= 0xffff;
   if ($code == 0) {
      return "exit code = 0";
   } elsif ($code == 0xff00) {
      return "command failed: $!"; 
   } elsif (($code & 0xff) == 0) {
      $code >>= 8;
      return "exit code = $code"; 
   } else {
      my $sig = ($code & ~0x80);
      if ($code & 0x80) {
         return "coredumped from signal $sig";
      } else {
         return "received signal $code";
      }
   }
}

my @files_to_delete;

sub
delete_files
{
   my $file;
   # delete files first then directories
   foreach $file (@files_to_delete) {
      unlink $file if ! -d $file;
   }
   foreach $file (@files_to_delete) {
      rmdir $file if -d $file;
   }
}

sub
make_tempfile()
{
   # Create a tempfile by first creating a mode 700 directory in
   # /tmp and then creating a file within that directory.  This
   # protects against the attack where someone predicts the name of
   # your tempfile and creates a symlink from that name to a file
   # that's writable by the user--- we open the tempfile, follow
   # the symlink and unwittingly truncate a file somewhere else in
   # the filesystem.  As long as /tmp has the sticky bit set on
   # it, we're safe from that attack by creating an intermediate
   # directory that only the user can write to.
   my $tempdir;
   my $n = 0;
   do {
      $tempdir = sprintf "/tmp/snap.%08d", rand(100_000_000);
   } while (! mkdir($tempdir, 0700) && ++$n < 100);

   if ($n == 100) {
      return undef;
   }

   push @files_to_delete, $tempdir;

   # make a tempfile to store the file we're building for rsync.
   my $tempfile = sprintf "$tempdir/snap.%08d.$$", rand(100_000_000);
   if (! open TEMP, ">$tempfile") {
      return undef;
   }
   close TEMP;

   push @files_to_delete, $tempfile;

   return $tempfile;
}
