# --include-- ../build/perl-invocation
# snap-backend - server that manages snap repositories
#
# Copyright (C) 2002 Kyle E. Jones
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program, in a file named COPYING.

# A snapshot is a directory containing a copy of a directory that
# exists somewhere else.  The source directory can live either in
# the local filesystem or a remote host.  Snapshot directory
# names are always the creation time of the snapshot expressed as
# the number of seconds since the epoch.
#
# A snapshot repository is a directory that contains snapshots.
# A repository directory also contains
#   - a symlink named 'oldest' that points to the oldest snapshot
#     in the repository
#   - a symlink named 'latest' which points to the youngest
#     snapshot in the repository.
#   - a file named 'source' which gives the host and directory
#     that the snapshot is a copy of.
#   - an optional file named 'encoder' which if present contains
#     a command to use as an encoder for regular files in the
#     snapshot.  E.g. "gzip -9" causes all the files in the
#     snapshot to be compressed.
#   - a file named 'snap-lock' which is used with flock() to
#     lock out other processes when the backend is modifying the
#     repository.

# --include-- common.pl
# --include-- ../build/system.pl

use File::Find;

my $hostname = `$hostname_program`;
chomp $hostname;

my $oldest_snapshot_name = 'oldest';
my $latest_snapshot_name = 'latest';
my $snapshot_lockfile = "snap-lock";

autoflush STDOUT 1;

# Create snapshot repositories and auxiliary files writable by
# owner and readable by others.  The snapshot directories
# themselves are created mode 700 initially.
umask 022;

# Commands arrive via stdin, we respond with this function on
# stdout.  First arg is typically the response code.  The rest
# of the args contain any data we want to pass back to the
# client.  All but the first arg is safety encoded, meaning
# control characters, and spaces are URL encoded.
sub
respond
{
   my $first = shift @_;
   print join(' ', $first, map { safety_encode($_) } @_), "\r\n";
}

# Die, sending an error message to the client.
sub
fatal
{
   respond @_;
   delete_files;
   exit 1;
}

# Read commands from stdin.
sub
read_command
{
   my $line = <STDIN>;
   $line =~ s/[\r\n]//g;
   return $line;
}

sub
prepare_directories_for_deletion
{
   my @dirs = @_;
   my $dir;

   foreach $dir (@dirs) {
      chmod 0700, $dir;
      opendir D, $dir;
      push @dirs, grep { $_ = "$dir/$_"; -d && ! /\/\.\.?$/ } readdir D;
      closedir D;
   }
}

# This function removes snapshots from a repository.  The current
# directory must be a snapshot repository at the time this function
# is called.
#
# First argument non-zero means don't actually delete any
# snapshots.  Instead send responses back the client detailing
# which snapshots would be kept and which would be deleted had
# this been a real pruning run.
#
# Second argument is a string naming the pruning method to use.
# Choices are 'keep-ages', 'prune-older' and 'snapshot'.  In the
# 'keep-ages' method the client provides the ages of snapshots
# to preserve.  For each age, the pruning function keeps the
# snapshot closest in age to each user specified age without
# being older.  If no younger snapshots are available for a
# particular age then the closest older snapshot is preserved.
# In the 'prune-older' method the client provides a single age
# and the pruning function deletes all the snapshots older than
# that age.  In the 'snapshot' method the provides the name of a
# snapshot to delete, or 'all' which means delete them all.
#
# For the 'snapshot' method the remaining argument is either a
# snapshot name or 'all'.  For the 'keep-ages and 'prune-older'
# methods, the remaining arguments are ages or ranges of ages.
# An age a string consisting of one or more digits followed by an
# optional single letter modifier.  Without a modifier the digits
# specify seconds.  With the modifier 's', 'm', 'h', 'd', and 'w'
# the digit specify seconds, minutes, hours, days and weeks
# respectively.  An age range is two sets of digits separates by
# a dash and optionally followed by a single letter modifier.
# Ranges specify a start age and an end age and are converted
# into an equivalent series of ages.  Example: the range 1-7d is
# equivalent to specifying the ages 1d 2d 3d 4d 5d 6d 7d.
sub
prune
{
   my ($dry_run, $method, @args) = @_;
   my $t;
   my (@newages, @ages);
   my %ages;
   if ($method eq 'prune-older' || $method eq 'keep-ages') {
      # Interpret the ages and ages ranges, storing them as a series
      # of ages specified in seconds.
      foreach $t (@args) {
	 # examples of legal age specs:
	 #   20
	 #   20d
	 #   2-20d
	 #   2d-20d
	 if ($t !~ /^([0-9]+)([smhdw]?-([0-9]+))?([smhdw]?)$/) {
	    respond "NO Bad age specification: $t";
	    return 0;
	 }
	 my $n = $1;
	 my $n2 = $3;
	 my $spec = $4;
	 $n2 = $n if $n2 eq '';
	 for (; $n <= $n2; $n++) {
	    my $total = $n;
	    if ($spec eq 'm') {
	       $total *= 60;
	    } elsif ($spec eq 'h') {
	       $total *= 60 * 60;
	    } elsif ($spec eq 'd') {
	       $total *= 24 * 60 * 60;
	    } elsif ($spec eq 'w') {
	       $total *= 7 * 24 * 60 * 60;
	    }
	    push @newages, $total;
	    $ages{$total} = $t;
	 }
      }
      # sort ages from most ancient to most recent
      @ages = sort { $b <=> $a } @newages;
   }
   # open the repository and get a list of all the snapshots.
   if (! opendir DOT, ".") {
      respond "NO couldn't opendir the current directory: $!";
      return 0;
   }
   my @dirs = grep /^[0-9]+$/, readdir DOT;
   closedir DOT;
   # sort snapshot directories from most ancient to most recent
   @dirs = sort { $a <=> $b } @dirs;
   my (%keep, @keep, %keepreason);
   if ($method eq 'prune-older' || $method eq 'keep-ages') {
      my @tempdirs = @dirs;
      my $now = time();
      # Start at the oldest age and work toward the youngest, marking
      # the closest snapshot in age to be kept (keep-ages) or
      # marking snapshots younger than the specified age to be kept
      # (prune-older).
      foreach $t (@ages) {
	 my $closest = -1;
	 my $closestdir = '';
	 my $dir;
	 if ($method eq 'keep-ages') {
	    while (defined($dir = shift @tempdirs)) {
	       if ($closest == -1) {
		  # use abs() because in this case $now - $t < $dir is possible
		  $closest = abs ($now - $dir - $t);
		  $closestdir = $dir;
	       } elsif ($now - $t < $dir) {
		  unshift @tempdirs, $dir;
		  last;
	       } elsif ($closest > $now - $dir - $t) {
		  $closest = $now - $dir - $t;
		  $closestdir = $dir;
	       }
	    }
	    if ($closestdir ne '') {
	       $keep{$closestdir} = 1;
	       $keepreason{$closestdir} = "matched by $ages{$t}";
	       push @keep, $closestdir;
	    }
	 } elsif ($method eq 'prune-older') {
	    my ($sec, $min, $hour, $mday, $mon, $year, undef)
	       = localtime($now - $t);
	    $year += 1900;
	    $mon += 1;
	    my $cutoffdate = sprintf "%d-%02d-%02d-%02d:%02d:%02d",
				     $year, $mon, $mday, $hour, $min, $sec;
	    foreach $dir (@tempdirs) {
	       next if $dir < $now - $t;
	       $keep{$dir} = 1;
	       $keepreason{$closestdir} = "younger than $cutoffdate";
	       push @keep, $dir;
	    }
	 }
      }
   } elsif ($method eq 'snapshot') {
      my $dir;
      if ($args[0] ne 'all') {
         foreach $dir (@dirs) {
            if ($args[0] ne $dir) {
	       $keep{$dir} = 1;
	       $keepreason{$dir} = "not specified for deletion";
	       push @keep, $dir;
	    }
         }
      }
   }
   # If this isn't a dry run, then update the 'oldest' and 'latest'
   # symlinks if we're keeping any snapshots.  If we aren't
   # keeping any snapshots, then just remove the two symlinks.
   if ($dry_run) {
      ; # do nothing
   } elsif ($keep[0] ne '') {
      my $x = 'oldest.' . $$;
      if (! symlink $keep[0], $x) {
         respond "NO symlink of $keep[0] to $x failed: $!";
         next COMMAND;
      }
      if (! rename $x, $oldest_snapshot_name) {
         unlink $x;
         respond "NO rename of $x to $oldest_snapshot_name failed: $!";
	 next COMMAND;
      }
      my $x = 'latest.' . $$;
      if (! symlink $keep[$#keep], $x) {
         respond "NO symlink of $keep[$#keep] to $x failed: $!";
         next COMMAND;
      }
      if (! rename $x, $latest_snapshot_name) {
         unlink $x;
         respond "NO rename of $x to $latest_snapshot_name failed: $!";
	 next COMMAND;
      }
   } else {
      unlink $latest_snapshot_name;
      unlink $oldest_snapshot_name;
   }
   # If we're doing a dry run, send the client a response for
   # each snapshot indicating whether we would keep the snapshot
   # or delete it.
   #
   # If we're doing a real run then delete the snapshots that are
   # not marked to be preserved.
   if ($dry_run) {
      my $dir;
      foreach $dir (@dirs) {
	 my ($sec, $min, $hour, $mday, $mon, $year, undef) = localtime($dir);
	 $year += 1900;
	 $mon += 1;
	 my $time = sprintf "%d-%02d-%02d-%02d:%02d:%02d",
			    $year, $mon, $mday, $hour, $min, $sec;
	 if ($keep{$dir}) {
            respond "KEEP $dir $time ($keepreason{$dir})";
         } else {
            respond "LOSE $dir $time";
         }
      }
   } else {
      my $dir;
      foreach $dir (@dirs) {
	 next if $keep{$dir};
	 if ($> != 0) { # if effective uid != 0
	    # Some directories might have the write bit turned off
	    # for the owner, which will prevent 'rm' from removing
	    # files in the directory if we're not running as the
	    # superuser.  Make all the directories writable by the
	    # owner before calling 'rm'.
	    prepare_directories_for_deletion $dir;
	 }
	 system "$rm_program", "-rf", "$dir";
	 if ($?) {
	    respond "NO rm -rf of $dir failed, ", quincy();
	    return 0;
	 }
      }
   }
   return 1;
}

# make a copy of a directory tree, creating matching symlinks for
# symlinks in the tree and hardlinking to everything else.  Copy
# owner, group, mode and access time information also.
sub
make_link_tree
{
   my ($sourcelink, $dest, $fullcopy) = (@_);
   my $source = $sourcelink;
   # Chase the top level symlink.  Currently we're always called
   # with a symlink as the first argument.
   while (-l $source) {
      $source = readlink $sourcelink;
      if ($source eq '') {
         respond "NO readlink of $sourcelink failed: $!";
         return 0;
      }
      $sourcelink = $source;
   }

   # make a list of the source directories
   if ($find_properties{accepts_dash_d}) {
      if (! open DIRS, "$find_program -d '$source' -type d |") {
	 respond "NO couldn't generate directory list: $!";
	 return 0;
      }
   } else {
      if (! open DIRS, "$find_program '$source' -depth -type d |") {
	 respond "NO couldn't generate directory list: $!";
	 return 0;
      }
   }
   my @dirs = <DIRS>;
   close DIRS;
   if ($?) {
      respond "NO a problem occurred generating the directory list (find), ",
	      quincy();
      return 0;
   }

   # Create the directories, copying owner, group and mode.
   # rsync will copy owner/group/mode if we don't, but if we do
   # it here then the rsync -v output will only show changed
   # files which will be useful to the user.
   my $i = $#dirs + 1;
   while ($i-- > 0) {
      my $dir = $dirs[$i];
      chomp $dir;
      next if $dir =~ m,/\.$,;
      next if $dir =~ m,/\.\.$,;
      my @foo = stat($dir);
      my $uid = $foo[4];
      my $gid = $foo[5];
      $dir =~ s,^${source},${dest},;
      # Create the directory mode 700 so the directory is initially
      # protected.
      if ($dir ne $dest && ! mkdir $dir, 0700) {
         respond "NO mkdir of $dir failed: $!";
         return 0;
      }
      chown $uid, $gid, $dir;
   }

   my $typearg;
   if ($fullcopy) {
      $typearg = '! -type d';
   } else {
      # only copy symlinks if we're not doing a full copy
      $typearg = '-type l';
   }
   # make a list of the source files
   if (! open FILES, "$find_program '$source' $typearg |") {
      respond "NO couldn't generate file list: $!";
      return 0;
   }

   # link the files into the directories
   # symlinks can't be linked to and so they must be created
   my $file;
   while (defined($file = <FILES>)) {
      chomp $file;
      my $old = $file;
      my $new = $file;
      $new =~ s,^${source},${dest},;
      if (-l $file) {
         # can't hard link to a symlink, so create a matching symlink
         # instead.
         $old = readlink $file;
         if ($old eq '') {
            respond "NO readlink of $file failed: $!";
            return 0;
         }
	 if (! symlink $old, $new) {
	    respond "NO symlink of $old -> $new failed: $!";
	    return 0;
	 }
      } elsif ($fullcopy) {
	 # hard link
	 if (! link $old, $new) {
	    respond "NO link of $old -> $new failed: $!";
	    return 0;
	 }
      }
   }
   close FILES;
   if ($?) {
      respond "NO a problem occurred generating the file list (find), ",
	      quincy();
      return 0;
   }

skip:
   # now copy the directory access times
   # again this is to make rsync -v only show the user real changes.
   my $i = $#dirs + 1;
   while ($i-- > 0) {
      my $dir = $dirs[$i];
      chomp $dir;
      next if $dir =~ m,/\.$,;
      next if $dir =~ m,/\.\.$,;
      my @foo = stat($dir);
      my $mode = $foo[2] & 0x0fff;
      my $atime = $foo[8];
      my $mtime = $foo[9];
      $dir =~ s,^${source},${dest},;
      chmod $mode, $dir;
      utime $atime, $mtime, $dir;
   }
   return 1;
}

my $cs_private_space;
my $cs_shared_space;
my $cs_ndirs;
my $cs_nfiles;
my %cs_xx_need_links;
my %cs_xx_got_links;
my %cs_xx_size;
my %cs_xx_mtime;

sub
add_some_disk_space
{
   my $file = $_;
   my ($dev, $ino, $mode, $nlink, $uid, $gid, $rdev, $size,
       $atime, $mtime, $ctime, $blksize, $blocks) = lstat($file);
   $cs_ndirs++ if -d _;
   $cs_nfiles++ if ! -d _;
   if (-l _ || -d _ || $nlink == 1) {
      $cs_private_space += $size;
   } else {
      my $x = "$dev/$ino";
      if (! $cs_xx_need_links{$x}) {
         $cs_xx_need_links{$x} = $nlink;
         $cs_xx_got_links{$x} = 1;
         $cs_xx_size{$x} = $size;
         $cs_xx_mtime{$x} = $mtime;
      } else {
         $cs_xx_got_links{$x} += 1;
      }
   }
}

sub
compute_snapshot_disk_usage
{
   my ($dir, $prevdir) = @_;

   %cs_xx_need_links = {};
   %cs_xx_got_links = {};
   %cs_xx_size = {};
   %cs_xx_mtime = {};
   $cs_private_space = 0;
   $cs_shared_space = 0;
   $cs_ndirs = 0;
   $cs_nfiles = 0;
   find(\&add_some_disk_space, $dir);
   my ($key, $value);
   while (($key, $value) = each %cs_xx_size) {
      if ($cs_xx_got_links{$key} == $cs_xx_need_links{$key}) {
         $cs_private_space += $cs_xx_size{$key};
      } elsif ($cs_xx_mtime{$key} > $prevdir) {
         $cs_shared_space += $cs_xx_size{$key};
      }
   }
   return ($cs_private_space, $cs_shared_space, $cs_ndirs, $cs_nfiles);
}

sub
read_repository_disk_usage
{
   if (! open DISKUSAGE, "<$diskusage_info_file") {
      return {};
   }
   my $line;
   my $usage = {};
   while (defined ($line = <DISKUSAGE>)) {
      chomp $line;
      my ($dir, $private, $shared, $ndirs, $nfiles) = split / +/, $line, 5;
      if ($dir eq '' || $private eq '' || $shared eq '') {
         respond "NO disk usage info file is corrupt";
         return undef;
      }
      $usage->{$dir}{private} = $private;
      $usage->{$dir}{shared} = $shared;
      $usage->{$dir}{ndirs} = $ndirs;
      $usage->{$dir}{nfiles} = $nfiles;
   }
   close DISKUSAGE;
   return $usage;
}

sub
write_repository_disk_usage
{
   my ($usage) = @_;
   my $workfile = sprintf "%d.%d", rand(10000), rand(10000);
   if (! open DISKUSAGE, ">$workfile") {
      respond "NO couldn't open $workfile (w): $!";
      return 0;
   }
   my $key;
   foreach $key (sort keys %{ $usage }) {
      my ($private, $shared, $ndirs, $nfiles) =
         ($usage->{$key}{private}, $usage->{$key}{shared},
	  $usage->{$key}{ndirs}, $usage->{$key}{nfiles});
      print DISKUSAGE "$key $private $shared $ndirs $nfiles\n";
   }
   if (! close DISKUSAGE) {
      respond "NO failed writing disk usage to $workfile: $!";
      return 0;
   }
   if (! rename $workfile, $diskusage_info_file) {
      respond "NO rename of $workfile to $diskusage_info_file failed: $!";
      return 0;
   }
   return 1;
}

sub
compute_repository_disk_usage
{
   my $usage = read_repository_disk_usage;

   if (! defined $usage) {
      # response already sent
      return undef;
   }

   # Walk through the disk usage list, clearing entries for
   # snapshots that no longer exist and invalidating entries for
   # snapshots that need to be recomputed.
   my $prevdir;
   my $dir;
   my $clear_next = 0;
   foreach $dir (sort keys %{ $usage }) {
      if (-d $dir) {
         # if snapshot entry no longer valid, clear it
         delete $usage->{$dir} if $clear_next;
	 $prevdir = $dir if ! $clear_next;
	 $clear_next = 0;
      } else {
	 # Because the snapshot for this entry no longer exists,
	 # bytes recorded as shared in the previous entry might
	 # now by private.  Or vice versa.  So the previous entry
	 # must be cleared so it will be recomputed later.
         delete $usage->{$prevdir} if defined $prevdir;
         # snapshot for this entry no longer exists, clear entry
	 delete $usage->{$dir};
	 undef $prevdir;
	 # Because the snapshot for this entry no longer exists,
	 # bytes recorded as shared in the next entry might now
	 # by private.  So the next entry (if there is one) must
	 # be cleared so it will be recomputed later.  Make a
	 # note to clear that entry.
	 $clear_next = 1;
      }
   }

   # open the repository and get a list of all the snapshots.
   if (! opendir DOT, ".") {
      respond "NO couldn't opendir the current directory: $!";
      return undef;
   }
   my @dirs = sort grep /^[0-9]+$/, readdir DOT;
   closedir DOT;

   # If the last usage entry is not the entry for the latest
   # snapshot, then the latest snapshot was created after this
   # entry was made.  Therefore the latest snapshot could now be
   # sharing some bytes from this snapshot that were marked as
   # private.  So that usage entry must be cleared as invalid.
   if (! defined $prevdir && $prevdir ne $dirs[$#dirs]) {
      delete $usage->{$prevdir};
   }

   # Now walk through the list of existing snapshots and compute
   # disk usage for those that do not already have usage entries
   # in $usage.
   $prevdir = 0;
   foreach $dir (@dirs) {
       if (! defined($usage->{$dir})) {
          # compute disk usage for this directory.
          my ($private, $shared, $ndirs, $nfiles) =
             compute_snapshot_disk_usage $dir, $prevdir;
	  $usage->{$dir}{private} = $private;
	  $usage->{$dir}{shared} = $shared;
	  $usage->{$dir}{ndirs} = $ndirs;
	  $usage->{$dir}{nfiles} = $nfiles;
       }
       $prevdir = $dir;
   }          

   # Store the usage information.
   if (! write_repository_disk_usage $usage) {
      # response already sent
      return undef;
   }

   return $usage;
}

sub
print_repository_disk_usage
{
   my $usage = compute_repository_disk_usage;
   return 0 if ! defined $usage;
   my $key;
   foreach $key (sort keys %{ $usage }) {
      my ($private, $shared, $ndirs, $nfiles) =
         ($usage->{$key}{private}, $usage->{$key}{shared},
	  $usage->{$key}{ndirs}, $usage->{$key}{nfiles});
      respond "USAGE", $key, $private, $shared, $ndirs, $nfiles;
   }
   return 1;
}

# lock a snapshot repository
sub
lock_repository
{
   if (! open SNAPLOCK, ">$snapshot_lockfile") {
      respond "NO couldn't open $snapshot_lockfile (w): $!";
      return 0;
   }
   if (! flock SNAPLOCK, $LOCK_EX|$LOCK_NB) {
      respond "NO couldn't flock snapshot lock: $!";
      return 0;
   }
   print SNAPLOCK @_, "\n";
   return 1;
}

# unlock a repository
sub
unlock_repository
{
   close SNAPLOCK;
}

# create and return a repository ID
sub
make_id
{
   return sprintf "SNAP-%04d", rand(10000);
}

sub
parse_command
{
   my ($line) = (@_);
   my (@args) = split / /, $line;
   my $first = shift @args;
   return ($first, map { safety_decode($_) } @args);
}

my $latest_snapshot_time_for_CHANGED;

sub
exit_0_if_changed
{
   my $foo = $_;
   # symlinks are created fresh so we can't compare their dates
   return if -l $foo;

   my $mtime = (stat($foo))[9];

   exit 0 if $mtime > $latest_snapshot_time_for_CHANGED;
}

### main program ###

# send greeting
respond "OK $hostname $progname ready";

my $line;
my $saw_start = 0;
my $saw_prepare = 0;

# Loop reading and executing commands from the client.
# Commands are:
#
#    REGISTER dir sourcedir - creates repository for sourcedir
#
#    UNREGISTER dir id - removes repository id in dir
#
#    CLEAN dir id - removes failed snapshots in repository id in dir
#
#    START dir id - cd's into dir and then id
#
#    ENCODER - returns an OK response containing the encoder that
#    was registered with this repository.  "-" is returned as the
#    encoder if this repository has no encoder.  START must be
#    issued before this command.
#
#    FIRST - returns an "OK YES" response if there are no snapshots
#    in this repository.  Otherwise an "OK NO" response is sent.
#    START must be issued before this command.
#
#    LINKDEST - returns an "OK YES" response if rsync on the machine
#    that is running that backend supports the --link-dest option.
#    Otherwise an "OK NO" response is sent.
#
#    DISKUSAGE - returns a series of USAGE responses containing
#    the disk usage of each snapshot in the repository.  The form
#    of the response is
#        USAGE snapshot_name private_bytes shared_bytes
#    private_bytes is the amount of disk space that would be
#    released if this snapshot were deleted.  shared_bytes is the
#    amount of disk space consumed by files that are shared among
#    multiple snapshots but was first stored in this snapshot.
#
#    PREPARE arg - creates a copy of the latest snapshot and
#    return the copy's name.  If arg is non-zero, do a full copy,
#    otherwise just copy the directory structure and symlinks.
#    START must be issued before this command.
#
#    CHANGED snapdir - returns a "OK YES" response if snapdir
#    contains files or directories with a modification time more
#    recent than the creation time of the most recent snapshot.
#    Otherwise return a "OK NO" response.
#
#    FINALIZE snapdir - moves a newly created snapshot from a temporary
#    to an official name, making it an official snapshot.  PREPARE
#    must be issued before this command.
#
#    SHOWPRUNE 'keep-ages|prune-older' age [ age ...]
#    SHOWPRUNE 'snapshot' snapshot | 'all' - reports which snapshots would
#    be removed if this were a PRUNE command, based on the pruning
#    method and specified args.
#
#    PRUNE 'keep-ages|prune-older' age [ age ...]
#    PRUNE 'snapshot' snapshot | 'all' - removes snapshots based
#    on the pruning method and specified args.
#
#    QUIT - ends the session

COMMAND: while (defined($line = read_command)) {
   if ($line =~ /^START/) {
      my (undef, $top_dir, $id) = parse_command($line);

      if ($top_dir eq '' || $id eq '') {
	 respond "NO Too few arguments.  Usage: START toplevel id";
	 next COMMAND;
      }
      if (! chdir $top_dir) {
         respond "NO", "chdir to $top_dir failed: $!";
         next COMMAND;
      }
      $saw_start = 0;
      if ($id =~ m,/, || $id eq '.' || $id eq '..') {
         respond "NO", "Bad ID: $id";
	 next COMMAND;
      }
      if (! chdir $id) {
         respond "NO", "chdir to $id failed: $!";
	 next COMMAND;
      }
      unlock_repository;
      if (! lock_repository) {
         # response has already been sent
         next COMMAND;
      }
      $saw_start = 1;
      respond "OK START $top_dir $id";
   } elsif ($line =~ /^REGISTER/) {
      my (undef, $top_dir, $sourcedir, $encoder) = parse_command($line);
      my $id;

      if ($top_dir eq '' || $sourcedir eq '' || $encoder eq '') {
	 respond "NO Too few arguments.  Usage: REGISTER toplevel sourcedir encoder";
      }
      if (! chdir $top_dir) {
         respond "NO", "chdir to $top_dir failed: $!";
         next COMMAND;
      }
      do {
         $id = make_id();
      } while (-e $id);
      if (! mkdir $id, 0777) {
         respond "NO", "mkdir of $id failed: $!";
	 next COMMAND;
      }
      if (! chdir $id) {
         respond "NO", "chdir to $id failed: $!";
         next COMMAND;
      }
      if (! open SOURCE, ">$source_info_file") {
         respond "NO", "couldn't open file in directory $id: $!";
         next COMMAND;
      }
      print SOURCE $sourcedir, "\n";
      if (! close SOURCE) {
         respond "NO", "couldn't write to file in directory $id: $!";
         next COMMAND;
      }
      if ($encoder ne '-') {
	 if (! open ENCODER, ">$encoder_info_file") {
	    respond "NO", "couldn't open file in directory $id: $!";
	    next COMMAND;
	 }
	 print ENCODER $encoder, "\n";
	 if (! close ENCODER) {
	    respond "NO", "couldn't write to file in directory $id: $!";
	    next COMMAND;
	 }
      }
      respond "OK", $id;
   } elsif ($line =~ /^UNREGISTER/) {
      my (undef, $top_dir, $id) = parse_command($line);

      if ($top_dir eq '' || $id eq '') {
	 respond "NO Too few arguments.  Usage: UNREGISTER toplevel id";
      }
      if ($id =~ m,/, || $id eq '.' || $id eq '..') {
         respond "NO", "Bad ID: $id";
	 next COMMAND;
      }
      if (! chdir $top_dir) {
         respond "NO", "chdir to $top_dir failed: $!";
         next COMMAND;
      }
      if ($> != 0) { # if effective uid != 0
	 # Some directories might have the write bit turned off
	 # for the owner, which will prevent 'rm' from removing
	 # files in the directory if we're not running as the
	 # superuser.  Make all the directories writable by the
	 # owner before calling 'rm'.
	 prepare_directories_for_deletion $id;
      }
      system "$rm_program", "-rf", "$id";
      if ($?) {
         respond "NO rm -rf of $id failed, ", quincy();
	 next COMMAND;
      }
      respond "OK $id unregistered, snapshots removed";
   } elsif ($line =~ /^CLEAN/) {
      my (undef, $top_dir, $id) = parse_command($line);

      if ($top_dir eq '' || $id eq '') {
	 respond "NO Too few arguments.  Usage: CLEAN toplevel id";
      }
      if ($id =~ m,/, || $id eq '.' || $id eq '..') {
         respond "NO", "Bad ID: $id";
	 next COMMAND;
      }
      if (! chdir $top_dir) {
         respond "NO", "chdir to $top_dir failed: $!";
         next COMMAND;
      }
      if (! chdir $id) {
         respond "NO", "chdir to $id failed: $!";
	 next COMMAND;
      }
      unlock_repository;
      if (! lock_repository) {
         # response has already been sent
         next COMMAND;
      }
      my $junkglob = '[0-9]*.[0-9]*';
      if ($> != 0) { # if effective uid != 0
	 # Some directories might have the write bit turned off
	 # for the owner, which will prevent 'rm' from removing
	 # files in the directory if we're not running as the
	 # superuser.  Make all the directories writable by the
	 # owner before calling 'rm'.
	 prepare_directories_for_deletion glob $junkglob;
      }
      my $cmd = "$rm_program -rf $junkglob";
      system $cmd;
      unlock_repository;
      if ($?) {
         respond "NO rm -rf of $junkglob failed, ", quincy();
	 next COMMAND;
      }
      respond "OK $id cleaned, abandoned working directories removed";
   } elsif ($line =~ /^ENCODER/) {
      respond "NO START command must be issued before this command"
         if ! $saw_start;
      if (! -f $encoder_info_file) {
	 respond "OK", "-";
	 next COMMAND;
      }
      if (! open ENCODER, "<$encoder_info_file") {
	 respond "NO", "couldn't open '$encoder_info_file': $!";
	 next COMMAND;
      }
      my $encoder = <ENCODER>;
      chomp $encoder;
      respond "OK", $encoder;
   } elsif ($line =~ /^LINKDEST/) {
      if (! system "$rsync_program -qn --link-dest=/ /tmp /tmp2 2>/dev/null") {
	 respond "OK", "YES";
	 next COMMAND;
      }
      respond "OK", "NO";
   } elsif ($line =~ /^FIRST/) {
      respond "NO START command must be issued before this command"
         if ! $saw_start;
      if (! -e $latest_snapshot_name) {
	 respond "OK", "YES";
	 next COMMAND;
      }
      respond "OK", "NO";
   } elsif ($line =~ /^DISKUSAGE/) {
      respond "NO START command must be issued before this command"
         if ! $saw_start;
      # We need to use find().  find() does chdir's and if it
      # fails it might leave us in a different directory.  So we
      # fork and run find() in the child.
      my $pid = fork;
      if ($pid) {
         # in parent
         waitpid $pid, 0;
         # child will respond to client
      } elsif (defined $pid) {
         # in child
         print_repository_disk_usage && respond "OK";
	 exit 0;
      } else {
         respond "NO", "fork failed: $!";
	 next COMMAND;
      }
   } elsif ($line =~ /^PREPARE/) {
      respond "NO START command must be issued before this command"
         if ! $saw_start;
      my (undef, $arg) = parse_command($line);
      $arg = 1 if ! defined $arg;
      my $name;
      do {
         $name = sprintf "%d.%d", time(), rand(100);
         # Create the directory mode 700 so the directory is
         # initially protected.
         if (! mkdir $name, 0700) {
            my $err = $!;
            if (! -e $name) {
               respond "NO", "mkdir of $name failed: $err";
	       next COMMAND;
            }
         }
      } while (! -d $name);
      if (-e $latest_snapshot_name) {
         my $result = make_link_tree($latest_snapshot_name, $name, $arg);
         if (! $result) {
            # make_link_tree already sent an error message to the
            # client.
            next COMMAND;
         }
      }
      $saw_prepare = 1;
      respond "OK", $name;
   } elsif ($line =~ /^FINALIZE/) {
      respond "NO PREPARE command must be issued before this command"
         if ! $saw_prepare;
      my (undef, $snapdir) = parse_command($line);
      if ($snapdir eq '') {
	 respond "NO Too few arguments.  Usage: FINALIZE dir";
	 next COMMAND;
      }
      my $new = $snapdir;
      $new =~ s,\..+$,,;
      if (! rename $snapdir, $new) {
         respond "NO", "rename of $snapdir to $new failed: $!";
	 next COMMAND;
      }
      if (! -e $oldest_snapshot_name) {
         if (! symlink $new, $oldest_snapshot_name) {
            respond "NO", "symlink of $new to $oldest_snapshot_name failed: $!";
            next COMMAND;
         }
      }
      my $latest = 'latest.' . $$;
      if (! symlink $new, $latest) {
         respond "NO", "symlink of $new to $latest failed: $!";
         next COMMAND;
      }
      if (! rename $latest, $latest_snapshot_name) {
         unlink $latest;
         respond "NO", "rename of $latest to $latest_snapshot_name failed: $!";
	 next COMMAND;
      }
      $saw_prepare = 0;
      respond "OK $new finalized as latest snapshot.";
   } elsif ($line =~ /^CHANGED/) {
      respond "NO PREPARE command must be issued before this command"
         if ! $saw_prepare;
      my (undef, $snapdir) = parse_command($line);
      if ($snapdir eq '') {
	 respond "NO Too few arguments.  Usage: CHANGED dir";
	 next COMMAND;
      }
      if (! -d $snapdir) {
         if (! -e $snapdir) {
	    respond "NO", "$snapdir does not exist";
         } else {
	    respond "NO", "$snapdir is not a directory";
	 }
	 next COMMAND;
      }
      # if there is no latest snapshot then rsync must have
      # changed something.
      if (! -d $latest_snapshot_name) {
         respond "OK", "YES";
         next COMMAND;
      }
      my $tempfile = make_tempfile();
      my $latest = readlink $latest_snapshot_name;;
      if (! defined $latest) {
         respond "NO", "readlink of $latest_snapshot_name failed: $!";
	 next COMMAND;
      }
      if ($latest !~ /^[0-9]+$/) {
         respond "NO", "format of $latest_snapshot_name link is not a number";
	 next COMMAND;
      }
      if (! utime $latest, $latest, $tempfile) {
         respond "NO", "utime of $tempfile failed: $!";
	 next COMMAND;
      }
      $latest_snapshot_time_for_CHANGED = $latest;
      # We need to use find().  find() does chdir's and if it
      # fails it might leave us in a different directory.  So we
      # fork and run find() in the child.
      my $pid = fork;
      if ($pid) {
         # in parent
         waitpid $pid, 0;
	 if (! $?) {
	    respond "OK", "YES";
	 } else {
	    respond "OK", "NO";
	 }
      } elsif (defined $pid) {
         # in child
         find(\&exit_0_if_changed, $snapdir);
	 exit 1;
      } else {
         respond "NO", "fork failed: $!";
	 next COMMAND;
      }
   } elsif ($line =~ /^SHOWPRUNE/) {
      respond "NO START command must be issued before this command"
         if ! $saw_start;
      my (undef, $method, @args) = parse_command($line);
      if ($method ne 'keep-ages' && $method ne 'prune-older' &&
          $method ne 'snapshot') {
         respond "NO unknown pruning method:", $method;
         next COMMAND;
      }
      if ($method eq 'prune-older' && defined $args[1]) {
         respond "NO only one age should be specified for pruning method",
		 $method;
      }
      if (@args == 0) {
         respond "NO no prune age or snapshot name specified";
         next COMMAND;
      }
      if (! prune(1, $method, @args)) {
         # response has already been sent
         next COMMAND;
      }
      respond "OK";
   } elsif ($line =~ /^PRUNE/) {
      respond "NO START command must be issued before this command"
         if ! $saw_start;
      my (undef, $method, @args) = parse_command($line);
      if ($method ne 'keep-ages' && $method ne 'prune-older' &&
          $method ne 'snapshot') {
         respond "NO unknown pruning method:", $method;
         next COMMAND;
      }
      if ($method eq 'prune-older' && defined $args[1]) {
         respond "NO only one age should be specified for pruning method",
		 $method;
      }
      if (@args == 0) {
         respond "NO no prune age or snapshot name specified";
         next COMMAND;
      }
      if (! prune (0, $method, @args)) {
         # response has already been sent
         next COMMAND;
      }
      respond "OK";
   } elsif ($line =~ /^QUIT/) {
      unlock_repository;
      respond "OK Bye";
      delete_files;
      exit 0;
   } else {
      my ($command, undef) = parse_command($line);
      respond "NO Unknown command: $command";
   }
}
