# --include-- ../build/perl-invocation
# snap-delete - remove specific snapshots
#
# Copyright (C) 2002 Kyle E. Jones
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program, in a file named COPYING.

# --include-- common.pl
# --include-- common-client.pl
# --include-- ../build/system.pl

sub
usage
{
   fatal
"usage: $progname [options] directory snapshot-name ...
        or
        $progname [options] directory all
options are:
   --verbose
   --registration-file=FILE
   --ssh-options=OPTIONS
";
}

my $ssh_options;

usage unless
   GetOptions(
	'verbose|debug+' => \$debug,
	'ssh-options=s' => \$ssh_options,
	'registration-file=s' => \$registration_file,
   );

my $sourcedir = shift @ARGV;
my @snapshots = @ARGV;

if ($sourcedir eq '') {
   diag "source directory missing";
   usage;
}

if (@snapshots == 0) {
   diag "snapshot to delete must be specified";
   usage;
}

if ($snapshots[0] eq 'all' && defined $snapshots[1]) {
   diag "either 'all' or a list of snapshots must be specified, not both";
   usage;
}

my $starting_cwd = cwd();
fatal "cwd() failed: $!" if $starting_cwd eq '';

if (! chdir $sourcedir) {
   fatal "couldn't chdir to $sourcedir: $!";
}

# canonicalize directory name;
$sourcedir = cwd();
fatal "cwd() failed: $!" if $sourcedir eq '';

chdir $starting_cwd ||
   fatal "couldn't cd back to $starting_cwd: $!";

# Search the registrations file for $sourcedir.
my $r = read_registrations;
fatal "$sourcedir: directory not registered" if ! defined $r->{$sourcedir};

my ($source, $id, $user, $remotehost, $remotedir) =
  ($r->{$sourcedir}{source},
   $r->{$sourcedir}{id},
   $r->{$sourcedir}{user},
   $r->{$sourcedir}{remotehost},
   $r->{$sourcedir}{repositories});

# Launch snap-backend, which will do the actual deletion of
# snapshots from the repository.
if ($remotehost ne '') {
   open2(\*FROMBACKEND, \*TOBACKEND,
         "$ssh_program $ssh_options -x -l $user $remotehost $snap_backend_program");
} else {
   open2(\*FROMBACKEND, \*TOBACKEND, "$snap_backend_program");
}

autoflush TOBACKEND 1;

# the deletions process:
# 1. Send the START command which tells the backend which
#    repository we'll be using.
# 2. Send the PRUNE command, which tells the backend which
#    snapshots to delete.  The pruning method is 'snapshot',
#    which means delete the snapshot we specify or if we
#    specifiy 'all' then delete all of them.

my $line = read_response;
fatal "no greeting from remote server" if $line eq '';
fatal "remote server greeted us with ", safety_decode($line) if $line !~ /^OK/;

send_command "START", $remotedir, $id;
$line = read_response;
fatal "START failed, remote server said ", safety_decode($line)
    if $line !~ /^OK/;

my $snapshot;
my $errors = 0;
foreach $snapshot (@snapshots) {
   # skip bogus snapshot names
   next if $snapshot !~ /^([0-9]+|all)$/;
   send_command "PRUNE", 'snapshot', $snapshot;
   $line = read_response;
   if ($line !~ /^OK/) {
      diag "PRUNE failed, remote server said ", safety_decode($line);
      $errors++;
   }
}

send_command "QUIT";
read_response;

exit $errors ? 1 : 0;
