# --include-- ../build/perl-invocation
# snap-snapshot - list the snapshots, giving some information about each
#
# Copyright (C) 2002 Kyle E. Jones
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program, in a file named COPYING.

# --include-- common.pl
# --include-- common-client.pl
# --include-- ../build/system.pl

sub
usage
{
   fatal
"usage: $progname [options] [directory ... | all ]
options are:
   --verbose
   --registration-file=FILE
   --ssh-options=OPTIONS
";
}

sub
S
{
   my ($count, $object_name) = @_;
   if ($count == 1) {
      return $object_name;
   } else {
      my $last_char = substr $object_name, -1, 1;
      if (length($object_name) == 1) {
         return $object_name . "'s";
      } elsif ($last_char eq 's') {
         return $object_name . "es";
      } elsif ($last_char eq 'y') {
        if (substr($object_name, -2, 1) =~ /[aeou]/) {
           return $object_name . "s";
        } else {
           return substr($object_name, 0, -1) . "ies";
        }
      } else {
         return $object_name . "s";
      }
   }
}

sub
prettybytes
{
   my ($size) = @_;
   my $KB = 1024;
   my $MB = ($KB * $KB);
   my $GB = ($MB * $KB);

   if ($size < $KB) {
      return "$size " . S($size, "byte");
   } elsif ($size < $MB) {
      # add 0.005 to round displayed value to the nearest hundredth
      return sprintf "%.2fKB", ($size / $KB) + 0.005;
   } elsif ($size < $GB) {
      return sprintf "%.2fMB", ($size / $MB) + 0.005;
   } else {
      return sprintf "%.2fGB", ($size / $GB) + 0.005;
   }
}

my $ssh_options;

usage unless
   GetOptions(
	'verbose|debug+' => \$debug,
	'ssh-options=s' => \$ssh_options,
	'registration-file=s' => \$registration_file,
   );

# read the registrations file
my $r = read_registrations;

if (@ARGV == 0) {
   push @ARGV, ".";
} elsif ($ARGV[0] eq 'all') {
   @ARGV = keys %{ $r };
}

my $starting_cwd = cwd();
fatal "cwd() failed: $!" if $starting_cwd eq '';

my ($sourcedir, $xxx);
my $errors = 0;
foreach $xxx (@ARGV) {

   $sourcedir = $xxx;

   # canonicalize the directory name
   if ($sourcedir eq '') {
      $sourcedir = cwd();
   } else {
      if (! chdir $sourcedir) {
	 diag "couldn't chdir to $sourcedir: $!";
	 diag "list of $xxx failed";
	 $errors++;
	 next;
      }
      $sourcedir = cwd();
   }
   if ($sourcedir eq '') {
      diag "cwd() failed: $!" if $sourcedir eq '';
      diag "list of $xxx failed";
      $errors++;
      next;
   }

   if (! chdir $starting_cwd) {
      diag "couldn't cd back to $starting_cwd: $!";
      $errors++;
      next;
   }

   # Search the registrations for $sourcedir or one of its
   # parent directories.  If there are multiple matches, choose the
   # one with the longest source path.
   my (@keys) = grep { subdirof($_, $sourcedir); } keys %{ $r };
   @keys = sort { length($b) <=> length($a) } @keys;
   my $key = $keys[0]; # after sort, longest is first in the array
   if ($key eq '') {
      diag "$sourcedir: directory not registered";
      diag "snapshot of $sourcedir failed";
      $errors++;
      next;
   }

   my ($source, $id, $user, $remotehost, $remotedir) =
     ($r->{$key}{source},
      $r->{$key}{id},
      $r->{$key}{user},
      $r->{$key}{remotehost},
      $r->{$key}{repositories});

   # Launch snap-backend, which manages the snapshot repository.
   if ($remotehost ne '') {
      open2(\*FROMBACKEND, \*TOBACKEND,
	    "$ssh_program $ssh_options -x -l $user $remotehost $snap_backend_program");
   } else {
      open2(\*FROMBACKEND, \*TOBACKEND, "$snap_backend_program");
   }

   autoflush TOBACKEND 1;

   # read server greeting
   my $line = read_response;
   if ($line eq '') {
      diag "no greeting from remote server";
      diag "list of $sourcedir failed";
      $errors++;
      close FROMBACKEND;
      close TOBACKEND;
      next;
   }
   if ($line !~ /^OK/) {
      diag "remote server greeted us with ", safety_decode($line);
      diag "list of $sourcedir failed";
      $errors++;
      close FROMBACKEND;
      close TOBACKEND;
      next;
   }

   # the listing process:
   # 1. Send the START command, telling the backend which snapshot
   #    repository we're going to be using.
   # 2. Send the DISKUSAGE command, telling the backend send us
   #    disk usage of about all the snapshots.  Format and print
   #    the results.

   send_command "START", $remotedir, $id;
   $line = read_response;
   if ($line !~ /^OK/) {
      diag "START failed, remote server said ", safety_decode($line);
      diag "list of $sourcedir failed";
      $errors++;
      goto shutdown_connection;
   }

   send_command "DISKUSAGE";
   my %usage;
   my %total;
   my $total = 0;
   my $max_private = 0;
   my $max_privshared = 0;
   my $max_cumbytes = 0;
   my $total_ndirs = 0;
   my $total_nfiles = 0;
   while (($line = read_response(1)) !~ /^(OK|NO)/) {
      if ($line =~ /^USAGE /) {
         $line = safety_decode($line);
         my (undef, $dir, $private, $shared, $ndirs, $nfiles) =
	    map { safety_decode($_) } split / /, $line, 6;
	 $total += $private + $shared;
	 $total_ndirs += $ndirs;
	 $total_nfiles += $nfiles;
         $usage{$dir}{private} = $private;
         $usage{$dir}{shared} = $shared;
         $total{$dir}{cumulative_bytes} = $total;

         $max_private = $private if $private > $max_private;
         $max_privshared = $private + $shared
            if $private + $shared > $max_privshared;
         $max_cumbytes = $total;
      }
   }
   if ($line !~ /^OK/) {
      diag "DISKUSAGE failed, remote server said ", safety_decode($line);
      diag "list of $sourcedir failed";
      $errors++;
      goto shutdown_connection;
   }

   print "$source repository $id";

   if ($remotehost ne '') {
      print " on ", $remotehost;
   }

   print " in $remotedir\n";
   print prettybytes($total), ", ",
	 $total_ndirs, " ", S($total_ndirs, "directory"), ", ",
	 $total_nfiles, " ", S($total_nfiles, "file"), "\n";

   goto shutdown_connection if ! %usage;

   print "\n";
   # generate a printf-style format with the minimal field width
   # needed to display each column.
   my $format = "%d %s";
   my $header1 = " snapshot     creation date   ";
   my $header2 = "---------- -------------------";
   my ($x, $columntitle);

   $columntitle = "private";
   $x = length($max_private);
   $x = length($columntitle) if $x < length($columntitle);
   $format .= " %${x}.0f";
   $columntitle = (' ' x (($x - length $columntitle) / 2)) . $columntitle;
   $columntitle .= (' ' x ($x - length $columntitle));
   $header1 .= " $columntitle";
   $header2 .= " " . "-" x $x;

   $columntitle = "combined";
   $x = length($max_privshared);
   $x = length($columntitle) if $x < length($columntitle);
   $format .= " %${x}.0f";
   $columntitle = (' ' x (($x - length $columntitle) / 2)) . $columntitle;
   $columntitle .= (' ' x ($x - length $columntitle));
   $header1 .= " $columntitle";
   $header2 .= " " . "-" x $x;

   $format .= " %6.2f%%";
   $header1 .= " % total";
   $header2 .= " -------";

   $columntitle = "cumul";
   $x = length($max_cumbytes);
   $x = length($columntitle) if $x < length($columntitle);
   $format .= " %${x}.0f";
   $columntitle = (' ' x (($x - length $columntitle) / 2)) . $columntitle;
   $columntitle .= (' ' x ($x - length $columntitle));
   $header1 .= " $columntitle";
   $header2 .= " " . "-" x $x;

   $format .= " %6.2f%%\n";
   $header1 .= " cumul %";
   $header2 .= " -------";

   print $header1, "\n";
   print $header2, "\n";

   my $key;
   foreach $key (sort keys %usage) {
      my ($sec, $min, $hour, $mday, $mon, $year, undef) = localtime($key);
      $year += 1900;
      $mon += 1;
      my $date = sprintf "%d-%02d-%02d-%02d:%02d:%02d",
			 $year, $mon, $mday, $hour, $min, $sec;
      printf $format,
	     $key, $date,
	     $usage{$key}{private},
	     $usage{$key}{private} + $usage{$key}{shared},
	     (($usage{$key}{private} + $usage{$key}{shared}) / $total) * 100,
	     $total{$key}{cumulative_bytes},
	     ($total{$key}{cumulative_bytes} / $total) * 100;
   }
   print "\n" if $xxx ne $ARGV[$#ARGV];

  shutdown_connection:
   send_command "QUIT";
   read_response;

   close FROMBACKEND;
   close TOBACKEND;
}

exit $errors ? 1 : 0;
