# --include-- ../build/perl-invocation
# snap-prune - remove some snapshots
#
# Copyright (C) 2002 Kyle E. Jones
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program, in a file named COPYING.

# --include-- common.pl
# --include-- common-client.pl
# --include-- ../build/system.pl

sub
usage
{
   fatal
"usage: $progname [options] directory age [age ...]
options are:
   --verbose
   --dry-run
   --registration-file=FILE
   --ssh-options=OPTIONS
";
}

my $ssh_options;
my $rsync_options;
my $dry_run = 0;

usage unless
   GetOptions(
	'verbose|debug+' => \$debug,
	'dry-run+' => \$dry_run,
	'ssh-options=s' => \$ssh_options,
	'registration-file=s' => \$registration_file,
   );

my $sourcedir = shift @ARGV;
my @ages = @ARGV;

if ($sourcedir eq '') {
   diag "source directory missing";
   usage;
}

if (@ages == 0) {
   diag "pruning ages missing";
   usage;
}

my $starting_cwd = cwd();
fatal "cwd() failed: $!" if $starting_cwd eq '';

if (! chdir $sourcedir) {
   fatal "couldn't chdir to $sourcedir: $!";
}

# canonicalize directory name
$sourcedir = cwd();
fatal "cwd() failed: $!" if $sourcedir eq '';

chdir $starting_cwd ||
   fatal "couldn't cd back to $starting_cwd: $!";

# Search the registrations file for $sourcedir.
my $r = read_registrations;
fatal "$sourcedir: directory not registered" if ! defined $r->{$sourcedir};

my ($source, $id, $user, $remotehost, $remotedir) =
  ($r->{$sourcedir}{source},
   $r->{$sourcedir}{id},
   $r->{$sourcedir}{user},
   $r->{$sourcedir}{remotehost},
   $r->{$sourcedir}{repositories});

# Launch snap-backend, which will do the actual deletion of
# snapshots from the repository.
if ($remotehost ne '') {
   open2(\*FROMBACKEND, \*TOBACKEND,
         "$ssh_program $ssh_options -x -l $user $remotehost $snap_backend_program");
} else {
   open2(\*FROMBACKEND, \*TOBACKEND, "$snap_backend_program");
}

autoflush TOBACKEND 1;

# the pruning process:
# 1. Send the START command which tells the backend which
#    repository we'll be using.
# 2. Send the PRUNE command, which tells the backend to look for
#    some snapshots to delete.  if we're doing a dry run, send
#    SHOWPRUNE instead.  The pruning method is 'keep-ages', which
#    means try to keep the snapshot closest (without being older)
#    in age to each of the specified ages.

my $line = read_response;
fatal "no greeting from remote server" if $line eq '';
fatal "remote server greeted us with ", safety_decode($line) if $line !~ /^OK/;

send_command "START", $remotedir, $id;
$line = read_response;
fatal "START failed, remote server said ", safety_decode($line)
    if $line !~ /^OK/;

my $prune_command = $dry_run ? "SHOWPRUNE" : "PRUNE";
send_command $prune_command, 'keep-ages', @ages;

if ($dry_run) {
   while (($line = read_response(1)) !~ /^(OK|NO)/) {
      if ($line =~ /^(KEEP|LOSE) /) {
         $line = safety_decode($line);
         print $line, "\n";
      }
   }
   fatal "$prune_command failed, remote server said ", safety_decode($line)
      if $line !~ /^OK/;
} else {
   $line = read_response;
   fatal "$prune_command failed, remote server said ", safety_decode($line)
      if $line !~ /^OK/;
}

send_command "QUIT";
read_response;

exit 0;
