# --include-- ../build/perl-invocation
# snap-register - register a directory for future snapshots
#
# Copyright (C) 2002 Kyle E. Jones
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program, in a file named COPYING.

# --include-- common.pl
# --include-- common-client.pl
# --include-- ../build/system.pl

sub
usage
{
   fatal
"usage: $progname [options] directory [[user@]remotehost:]repositories-dir
options are:
   --verbose
   --encoder=COMMAND
   --registration-file=FILE
   --ssh-options=OPS
";
}

my $ssh_options;
my $encoder;

usage unless
   GetOptions(
	'verbose|debug+' => \$debug,
	'encoder=s' => \$encoder,
	'registration-file=s' => \$registration_file,
	'ssh-options=s' => \$ssh_options,
   );

my $sourcedir = $ARGV[0];
my $repositories_dir = $ARGV[1];

if ($sourcedir eq '') {
   diag "source directory missing";
   usage;
}
if ($repositories_dir eq '') {
   diag "repositories directory missing";
   usage;
}

# second arg can be just a directory or can be of the form
#
# [[user@]remotehost:]repositories-dir
#
# which we must parse.
my ($remotehost, $dir);
if ($repositories_dir =~ /:/) {
   ($remotehost, $dir) = split /:/, $repositories_dir, 2;
   if ($dir ne '') {
      $repositories_dir = $dir;
      my $u;
      ($u, $remotehost) = split /\@/, $remotehost, 2;
      if ($remotehost eq '') {
         $remotehost = $u;
      } else {
         $user = $u
      }
   } else {
      $remotehost = '';
   }
}

fatal "who are you?" if ($user eq '');

my $starting_cwd = cwd();
fatal "cwd() failed: $!" if $starting_cwd eq '';

if (! chdir $sourcedir) {
   fatal "couldn't chdir to $sourcedir: $!";
}

# canonicalize directory name;
$sourcedir = cwd();
fatal "cwd() failed: $!" if $sourcedir eq '';

chdir $starting_cwd ||
   fatal "couldn't cd back to $starting_cwd: $!";

# read the registration file and lock it since we are going to
# change it.
my $r = read_and_lock_registrations;

fatal "$sourcedir: directory already registered" if defined $r->{$sourcedir};

# Launch snap-backend, which will create the snapshot repository.
if ($remotehost ne '') {
   open2(\*FROMBACKEND, \*TOBACKEND,
         "$ssh_program $ssh_options -x -l $user $remotehost $snap_backend_program");
} else {
   open2(\*FROMBACKEND, \*TOBACKEND, "$snap_backend_program");
}

autoflush TOBACKEND 1;

# the registration process:
# 1. Send the REGISTER command.  The backend will create the
#    repository and respond a repository ID, which will be
#    stored in the registrations file to be used in the future.

my $line = read_response;
fatal "no greeting from remote server" if $line eq '';
fatal "remote server greeted us with ", safety_decode($line) if $line !~ /^OK/;

my $hostname = `$hostname_program`;
chomp $hostname;

send_command "REGISTER", $repositories_dir, "$hostname:$sourcedir",
	     defined($encoder) ? $encoder : '-';
$line = read_response;
fatal "REGISTER failed, remote server said ", safety_decode($line)
   if $line !~ /^OK/;
my (undef, $id) = map { safety_decode($_) } split / /, $line, 2;
fatal "ID missing in REGISTER response" if $id eq '';

# add the new registration to $r
$r->{$sourcedir}{source} = $sourcedir;
$r->{$sourcedir}{id} = $id;
$r->{$sourcedir}{user} = $user;
$r->{$sourcedir}{remotehost} = $remotehost;
$r->{$sourcedir}{repositories} = $repositories_dir;

# write out the registrations including the new one
write_registrations($r);

send_command "QUIT";
read_response;
exit 0;
