# --include-- ../build/perl-invocation
# snap-snapshot - recreate a directory elsewhere, copying minimally
#
# Copyright (C) 2002 Kyle E. Jones
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program, in a file named COPYING.

# --include-- common.pl
# --include-- common-client.pl
# --include-- ../build/system.pl

sub
usage
{
   fatal
"usage: $progname [options] [directory ...]
options are:
   --verbose
   --if-changed
   --registration-file=FILE
   --exclusion-file=FILE
   --rsync-options=OPTIONS
   --ssh-options=OPTIONS
";
}

# Read the exclusions/inclusions for a directory and store them
# in an exclusions file suitable for use by rsync.
#
# The exclusion file consists of three types of lines:
#
# 1. Junk.  Lines that begin with # or lines that conists of all
#    whitespace.  These lines are ignored.
# 2. Lines of the form "directory FOO", which specify the directory to
#    which the following exclusions and inclusions apply.  Exclusions
#    and inclusions on subsequent lines apply only to snapshots of
#    directory FOO until another "directory" line is encountered.
# 3. Lines that being with "+ " or "- ".  These specify inclusions
#    and exclusions respectively, with semantics exactly matching
#    those described in the rsync(1) man page, EXCEPT that to make
#    inclusions more convenient to use, all parent directories on
#    an inclusion that contains a / are automatically included.
#    Example:
#
#	+ /usr/local/bin/perl
#
#    is equivalent to writing
#
#	+ /usr/local/bin/perl
#	+ /usr/local/bin/
#	+ /usr/local/
#	+ /usr/
#	+ /
#
#    which is what you would need to write to get the desired
#    effect.  The inclusions apply to snapshots of the directory
#    specified in the most recent "directory" declaration line,
#    or to all snapshots if no "directory" declarations have yet
#    been encountered in the file.

sub
read_exclusions
{
   my ($directory) = @_;

   # the exclusion file is optional so just return if none is defined.
   return undef if ! defined $exclusion_file;

   # croak if an exclusion file is defined but does not exist
   if (! open EX, "<$exclusion_file") {
      diag "couldn't open exclusion file $exclusion_file (r): $!";
      return "";
   }

   my $tempfile = make_tempfile();
   if (! defined $tempfile) {
      diag "couldn't create a tempfile with make_tempfile()";
      close EX;
      return "";
   }

   if (! open TEMP, ">$tempfile") {
      diag "couldn't open $tempfile (w): $!";
      close EX;
      return "";
   }

   my $line;
   my $lineno = 0;
   my $curdir;
   my %included_dirs;
   while (defined($line = <EX>)) {
      chomp $line;
      $lineno++;
      # skip junk
      next if $line =~ /^[ \t]*$/;
      next if $line =~ /^#/;
      # all other valid lines have exactly two fields
      my ($op, $arg) = split /[ \t]+/, $line, 2;
      # decode the second field
      $arg = safety_decode($arg);
      if ($op eq 'directory') {
	 # directory declaration
         $curdir = $arg;
      } elsif ($arg eq '') {
	 # syntax error: all lines must contain two field
         diag "$exclusion_file: line $lineno: too few fields";
	 close EX;
	 close TEMP;
	 return "";
      } elsif (defined $curdir && $curdir ne $directory) {
         ; # exclusion/inclusion not for the directory we're reading for
      } elsif ($op ne '-' && $op ne '+') {
	 # unknown exclusion operator
         diag "$exclusion_file: line $lineno: bad operator '$op'";
	 close EX;
	 close TEMP;
	 return "";
      } else {
	 # passed all the tests, write it out
         print TEMP "$op $arg\n";
	 # For inclusion lines, make them more convenient by
	 # automatically including any parent directories.
         if ($op eq '+' && $arg =~ m,/,) {
	    $arg =~ s,/$,,g;
	    $arg =~ s,/[^/]+$,/,;
	    while ($arg =~ m,/,) {
	       if (! $included_dirs{$arg}) {
	          print TEMP "+ $arg\n";
	          $included_dirs{$arg} = 1;
               }
	       $arg =~ s,/$,,g;
	       $arg =~ s,/[^/]+$,/,;
	    }
         }
      }
   }

   close EX;

   # catch any write errors here
   if (! close TEMP) {
      diag "close of $tempfile failed: $!";
      return "";
   }

   # If we created the tempfile but didn't write anything, return
   # undef instead of a filename pointing to a useless empty
   # file.
   if (! -s $tempfile) {
      return undef;
   }

   return $tempfile;
}

my $ssh_options;
my $rsync_options;
my $if_changed = 0;

usage unless
   GetOptions(
	'verbose|debug+' => \$debug,
	'if-changed+' => \$if_changed,
	'ssh-options=s' => \$ssh_options,
	'rsync-options=s' => \$rsync_options,
	'registration-file=s' => \$registration_file,
	'exclusion-file=s' => \$exclusion_file,
   );

if (@ARGV == 0) {
   push @ARGV, ".";
}

# read the registrations file
my $r = read_registrations;

my $starting_cwd = cwd();
fatal "cwd() failed: $!" if $starting_cwd eq '';

my ($sourcedir, $xxx);
my $errors = 0;
foreach $xxx (@ARGV) {

   $sourcedir = $xxx;

   # canonicalize the directory name
   if ($sourcedir eq '') {
      $sourcedir = cwd();
   } else {
      if (! chdir $sourcedir) {
	 diag "couldn't chdir to $sourcedir: $!";
	 diag "snapshot of $xxx failed";
	 $errors++;
	 next;
      }
      $sourcedir = cwd();
   }
   if ($sourcedir eq '') {
      diag "cwd() failed: $!" if $sourcedir eq '';
      diag "snapshot of $xxx failed";
      $errors++;
      next;
   }

   if (! chdir $starting_cwd) {
      diag "couldn't cd back to $starting_cwd: $!";
      diag "snapshot of $xxx failed";
      $errors++;
      next;
   }

   # Search the registrations for $sourcedir or one of its
   # parent directories.  If there are multiple matches, choose the
   # one with the longest source path.
   my (@keys) = grep { subdirof($_, $sourcedir); } keys %{ $r };
   @keys = sort { length($b) <=> length($a) } @keys;
   my $key = $keys[0]; # after sort, longest is first in the array
   if ($key eq '') {
      diag "$sourcedir: directory not registered";
      diag "snapshot of $sourcedir failed";
      $errors++;
      next;
   }

   diag "making snapshot of $sourcedir" if $debug;

   my ($source, $id, $user, $remotehost, $remotedir) =
     ($r->{$key}{source},
      $r->{$key}{id},
      $r->{$key}{user},
      $r->{$key}{remotehost},
      $r->{$key}{repositories});

   # get the exclusions/inclusions for this directory
   my $exclusions = read_exclusions($source);
   if (defined $exclusions && $exclusions eq '') {
      # read_exclusions already wrote an error message
      diag "snapshot of $sourcedir failed";
      $errors++;
      last;
   }

   # Launch snap-backend, which manages the snapshot repository.
   if ($remotehost ne '') {
      open2(\*FROMBACKEND, \*TOBACKEND,
	    "$ssh_program $ssh_options -x -l $user $remotehost $snap_backend_program");
   } else {
      open2(\*FROMBACKEND, \*TOBACKEND, "$snap_backend_program");
   }

   autoflush TOBACKEND 1;

   # read server greeting
   my $line = read_response;
   if ($line eq '') {
      diag "no greeting from remote server";
      diag "snapshot of $sourcedir failed";
      $errors++;
      close FROMBACKEND;
      close TOBACKEND;
      next;
   }
   if ($line !~ /^OK/) {
      diag "remote server greeted us with ", safety_decode($line);
      diag "snapshot of $sourcedir failed";
      $errors++;
      close FROMBACKEND;
      close TOBACKEND;
      next;
   }

   # the snapshot process:
   # 1. Send the START command, telling the backend which snapshot
   #    repository we're going to be using.
   # 2. Send the PREPARE command, telling the backend to create a
   #    copy of the latest snapshot for us to rsync against.  The
   #    backend responds with the name of this directory containing
   #    this copy.
   # 3. rsync the source dir to the snapshot copy, which will replace
   #    files in the copy that have changed in the source directory.
   # 4. Send the FINALIZE command, which tells the backend to install
   #    the snapshot copy as an official snapshot in the repository.

   send_command "START", $remotedir, $id;
   $line = read_response;
   if ($line !~ /^OK/) {
      diag "START failed, remote server said ", safety_decode($line);
      diag "snapshot of $sourcedir failed";
      $errors++;
      goto shutdown_connection;
   }

   send_command "ENCODER";
   $line = read_response;
   if ($line !~ /^OK/) {
      diag "ENCODER failed, remote server said ", safety_decode($line);
      diag "snapshot of $sourcedir failed";
      $errors++;
      goto shutdown_connection;
   }
   my (undef, $encoder) = map { safety_decode($_) } split / /, $line, 2;
   $encoder = undef if $encoder eq '-';

   my $link_dest_supported = 0;
   if (! system "$rsync_program -qn --link-dest=/ /tmp /tmp2 2>/dev/null") {
      $link_dest_supported = 1;
      # no need to do the remote test if we're everything is local.
      unless ($remotehost eq '') {
	  send_command "LINKDEST";
	  $line = read_response;
	  if ($line !~ /^OK/) {
	     diag "LINKDEST failed, remote server said ", safety_decode($line);
	     diag "snapshot of $sourcedir failed";
	     $errors++;
	     goto shutdown_connection;
	  }
	  (undef, $link_dest_supported) =
	     map { safety_decode($_) } split / /, $line, 2;
	  $link_dest_supported = ($link_dest_supported eq 'YES') ? 1 : 0;
      }
   }

   my $first_snapshot = 0;
   if ($link_dest_supported) {
      send_command "FIRST";
      $line = read_response;
      if ($line !~ /^OK/) {
	 diag "FIRST failed, remote server said ", safety_decode($line);
	 diag "snapshot of $sourcedir failed";
	 $errors++;
	 goto shutdown_connection;
      }
      (undef, $first_snapshot) =
         map { safety_decode($_) } split / /, $line, 2;
      $first_snapshot = ($first_snapshot eq 'YES') ? 1 : 0;
   }

   send_command "PREPARE", ($first_snapshot || ! $link_dest_supported) ? 1 : 0;
   $line = read_response;
   if ($line !~ /^OK/) {
      diag "PREPARE failed, remote server said ", safety_decode($line);
      diag "snapshot of $sourcedir failed";
      $errors++;
      goto shutdown_connection;
   }
   my (undef, $dir) = map { safety_decode($_) } split / /, $line, 2;
   if ($id eq '') {
      diag "directory missing in PREPARE response";
      diag "snapshot of $sourcedir failed";
      $errors++;
      goto shutdown_connection;
   }

   # Now rsync the source directory
   if ($remotehost ne '') {
      my @args = ("$rsync_program", "--rsh", "$ssh_program",
		  "--recursive", "--perms", "--times",
		  "--hard-links", "--links", "--delete");
      # can't reliably maintain group/owner unless we're root.
      # assumption here is that if we're root here, we're root on
      # the remote side also.
      if ($> == 0) {
         push @args, "--owner", "--group";
      }
      if ($> == 0 && $link_dest_supported && ! $first_snapshot) {
         push @args, "--link-dest", "$remotedir/$id/latest/.";
      }
      if (defined $exclusions) {
	 push @args, "--delete-excluded", "--exclude-from", $exclusions;
      }
      if (defined $encoder) {
         push @args, "--times-only", "--dest-filter", $encoder;
      }
      if ($debug) {
	 push @args, "--verbose";
      }
      if (defined $rsync_options) {
	 push @args, split /[ \t]+/, $rsync_options;
      }
      push @args, "$source/", "$user\@$remotehost:$remotedir/$id/$dir";
      diag "rsyncing data to $user\@$remotehost:$remotedir/$id/$dir" if $debug;
      system @args;
   } else {
      my @args = ("$rsync_program", "--rsh", "$ssh_program",
		  "--recursive", "--perms", "--times",
		  "--hard-links", "--links", "--delete");
      # can't reliably maintain group/owner unless we're root
      if ($> == 0) {
         push @args, "--owner", "--group";
      }
      if ($> == 0 && $link_dest_supported && ! $first_snapshot) {
         push @args, "--link-dest", "$remotedir/$id/latest/.";
      }
      if (defined $exclusions) {
	 push @args, "--delete-excluded", "--exclude-from", $exclusions;
      }
      if (defined $encoder) {
         push @args, "--times-only", "--dest-filter", $encoder;
      }
      if ($debug) {
	 push @args, "--verbose";
      }
      if (defined $rsync_options) {
	 push @args, split /[ \t]+/, $rsync_options;
      }
      push @args, "$source/", "$remotedir/$id/$dir";
      diag "rsyncing data to $remotedir/$id/$dir" if $debug;
      system @args;
   }
   if ($?) {
      diag "rsync failed, ", quincy();
      diag "snapshot of $sourcedir failed";
      $errors++;
      goto shutdown_connection;
   }
   diag "rsync complete" if $debug;

   if ($if_changed) {
      send_command "CHANGED", $dir;
      $line = read_response;
      if ($line !~ /^OK/) {
	 diag "CHANGED failed, remote server said ", safety_decode($line);
	 diag "snapshot of $sourcedir failed";
	 $errors++;
	 goto shutdown_connection;
      }
      if ($line =~ /NO/) {
         diag "$sourcedir unchanged, will not retain snapshot" if $debug;
	 send_command "CLEAN", $remotedir, $id;
	 $line = read_response;
	 goto shutdown_connection;
      }
   }
   send_command "FINALIZE", $dir;
   $line = read_response;
   if ($line !~ /^OK/) {
      diag "FINALIZE failed, remote server said ", safety_decode($line);
      diag "snapshot of $sourcedir failed";
      $errors++;
      goto shutdown_connection;
   }

  shutdown_connection:
   send_command "QUIT";
   read_response;

   close FROMBACKEND;
   close TOBACKEND;
}

delete_files;

exit $errors ? 1 : 0;
