# --include-- ../build/perl-invocation
# snap-subview - create a user-friendly view of the snapshots of a directory
#
# Copyright (C) 2002 Kyle E. Jones
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program, in a file named COPYING.

# --include-- common.pl
# --include-- common-client.pl
# --include-- ../build/system.pl

sub
usage
{
   fatal
"usage: $progname [options] view-directory [directory]
options are:
   --registration-file=FILE
   --repositories-directory=DIRECTORY
";
}

my $repositories_directory;

usage unless
   GetOptions(
	'registration-file=s' => \$registration_file,
	'repositories-directory=s' => \$repositories_directory,
   );

my $starting_cwd = cwd();
fatal "cwd() failed: $!" if $starting_cwd eq '';

my $subviewdir = $ARGV[0];
if ($subviewdir eq '') {
   diag "view-directory missing";
   usage;
}

my $dirtoview = $ARGV[1];
if ($dirtoview ne '') {
   chdir $dirtoview or
      fatal "couldn't cd to $dirtoview: $!";
}

# canonicalize the directory name
my $canon_dtv = cwd();
fatal "cwd() failed after chdir to $dirtoview: $!" if $canon_dtv eq '';

chdir $starting_cwd ||
   fatal "couldn't cd back to $starting_cwd: $!";

my $domain = `$hostname_program`;
chomp $domain;
$domain = lc($domain);
$domain =~ s/^[^.]+\.//;

# Search the registrations file for $canon_dtv or one of its
# parent directories.  If there are multiple matches, choose the
# one with the longest source path.
my $r = read_registrations;
my (@keys) = grep { subdirof($_, $canon_dtv); } keys %{ $r };
@keys = sort { length($b) <=> length($a) } @keys;
my $key = $keys[0]; # after sort, longest is first in the array
fatal "couldn't find snapshot registration for any parents of this directory"
   if $key eq '';

# Now find the repository.  If the repositories directory was
# given on the command line, use that.  Else if the repository
# isn't on a remote host, it must be a local filesystem, so use
# the repositories directory that is sepcfieid in the reigstration
# entry.  Else the repositories directory must be on a remote
# host.  In that case use the 'mount' command to generate a list
# of mount filesystems and try to find the mount point of the
# remote repositories or a parent directory.
my $repository;
if ($repositories_directory ne '') {
   # use command line info
   $repository = $repositories_directory .  "/" . $r->{$key}{id};
} elsif ($r->{$key}{remotehost} eq '') {
   # use info from the snapshot registration
   $repository = $r->{$key}{repositories} . "/" . $r->{$key}{id};
} else {
   # search for the repositories among mounted filesystems
   open MOUNT, "$mount_program|" or
      fatal "couldn't run $mount_program: $!";
   my $line;
   while (defined ($line = <MOUNT>)) {
      # The syntax we expect is "SPECIAL on MOUNTPOINT ...".
      # Unfortunately some UNIXes reverse SPECIAL and MOUNTPOINT.
      # We deal with this by checking $special to see if it is a
      # directory.  If it is, then we swap $special and $mountpoint.
      my ($special, $undef, $mountpoint, undef) = split / /, $line;
      if (-d $special) {
         my $tmp = $special;
	 $special = $mountpoint;
	 $mountpoint = $tmp;
      }
      # we're looking for NFS mounts, so skip SPECIALs that don't
      # contain a colon as these are local devices.
      next if $special !~ /:/;
      my ($host, $dir) = split /:/, $special, 2;
      if (hostmatch ($host, $r->{$key}{remotehost}, $domain) &&
	  subdirof($dir, $r->{$key}{repositories})) {
	 # Found a match.  Start with the remote path to the repositories.
         $repository = $r->{$key}{repositories};
	 # Replace the mounted remote directory that occurs
	 # within that path with the local mountpoint.
	 $repository =~ s,^${dir},${mountpoint},;
	 # Append the repository ID to complete the path.
	 $repository = $repository . "/" . $r->{$key}{id};
         last;
      }
   }
   fatal "couldn't find snapshots of $r->{$key}{source} mounted anywhere"
      if $repository eq '';
}

# Now open the repository directory.
opendir REPOSITORY, $repository or
   fatal "couldn't opendir $repository: $!";

# All snapshot names are a string of digits.  Only snapshots have
# names of this format.
my @snapshots = grep /^[0-9]+$/, readdir REPOSITORY;
closedir REPOSITORY;

if (! @snapshots) {
   fatal "there are no snapshots of $r->{$key}{source}";
}

if (-l $subviewdir) {
   fatal "$subviewdir is a symlink; must either not exist or be a directory";
}

# create the view directory if it doesn't already exist.
if (! -d $subviewdir && ! mkdir $subviewdir, 0777) {
   fatal "couldn't mkdir $subviewdir: $!";
}

# For each snapshot, create a symlink to the snapshot of the
# current directory within that snapshot.
my $errors = 0;
my $snap;
foreach $snap (@snapshots) {
   my ($sec, $min, $hour, $mday, $mon, $year, undef) = localtime($snap);
   $year += 1900;
   $mon += 1;
   my $link = sprintf "%d-%02d-%02d-%02d:%02d:%02d",
		      $year, $mon, $mday, $hour, $min, $sec;
   # Create path to the snapshot of the source directory.
   # Start with the source directory.
   my $relpath = $canon_dtv;
   # Strip off the source directory to a relative path.
   $relpath =~ s,^$r->{$key}{source}/*,,;
   # Concatenate the path to the snapshot with the relative path
   # to produce the path to the snapshot of $canon_dtv.
   my $old = "$repository/$snap/$relpath";
   # The source directory may not have existed in all the snapshots so
   # check for its existence in this snapshot before creating the link.
   next if ! -d $old;
   my $new = "$subviewdir/$link";
   if (! -l $new && ! symlink $old, $new) {
      diag "symlink of $old to $new failed: $!";
      $errors++;
   }
}

# Now look for symlinks in $subviewdir that point to directories
# that are no longer in the filesystem.  Any such links must be
# pointing at directories that have been removed by snap-prune or
# snap-baseline.

# Single quotes are around $subviewdir to avoid problems if it
# contains spaces.
if (! open LINKS,
           "$find_program '$subviewdir' -type l -name '*-??-??-??:??:??' |") {
   fatal "couldn't generate symlink list with $find_program: $!";
}

my $link;
while (defined ($link = <LINKS>)) {
   chomp $link;
#   unlink $link if ! -d $link;
}

close LINKS;

exit 1 if $errors;
exit 0;
