# --include-- ../build/perl-invocation
# snap-view - create a user-friendly tree view of snapshot repositories
#
# Copyright (C) 2002 Kyle E. Jones
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program, in a file named COPYING.

# --include-- common.pl
# --include-- common-client.pl
# --include-- ../build/system.pl

sub
usage
{
   fatal "usage: $progname view-directory snapshot-repository [snapshot-repository ...]";
}

my $viewdir = $ARGV[0];
if ($viewdir eq '') {
   diag "view-directory missing";
   usage;
}

shift @ARGV;
if ($ARGV[0] eq '') {
   diag "at least one snapshot directory must be given";
   usage;
}

# Get canonical name of current directory.  This is used to
# produce full paths if the snapshot directoriues on the commadn
# line are relative paths.
my $cwd = cwd();
fatal "cwd() failed: $!" if $cwd eq '';

# Extract the domain part of the hostname.  This domain is trimmed
# off hostnames that share the same domain.
my $domain_regexp = `$hostname_program`;
chomp $domain_regexp;  # strip newline
$domain_regexp = lc($domain_regexp);  # lower
$domain_regexp =~ s/^[^.]+\.//;  # foo.bar.com -> bar.com
$domain_regexp =~ s/\./\\./g; # quote dots

# @ARGV contains snapshot repositories.  These are directories
# with names of the form SNAP-NNNN, where NNNN is a number.  Each
# repository contains snapshot directories whose names are their
# creation times expresed as the number of seconds since the
# epoch.  There is also a file called "source" which gives the
# hostname and directory from which the snapshot originated.
#
# We are building a directory tree rooted at $viewdir.  There is
# a path from the top of the tree to each snapshot directory in
# the snapshot repositories.  The path looks like
#
#     $viewdir/HOSTNAME/SNAP-PATH/TIME
#
# HOSTNAME is the source host of the snapshot.
#
# SNAP-PATH is the full path of the source directory corresponding
# to the snapshot with the leading slash removed and the remaining
# slashes converted to dashes.  E.g. /usr/home is usr-home.  For
# the special case of /, SNAP-PATH is -ROOT-.
#
# TIME is the time the snapshot was created, in ISO 8601 standard
# date format.  TIME is a symlink to the corresponding snapshot
# directory.

# Create the top of the tree, reusing a directory if it already
# exists.  This is important because processes may already have
# this directories or subdirectories of it as their current
# directory.  It would be bad to remove the directory from under
# them.
if (-l $viewdir) {
   fatal "$viewdir is a symlink; must either not exist or be a directory";
}
if (! -d $viewdir && ! mkdir $viewdir, 0777) {
   fatal "couldn't mkdir $viewdir: $!";
}

my $repository;
my $errors = 0;
REPOSITORY: foreach $repository (@ARGV) {
   # Get the source information
   # File should contain one line of the form
   #    HOST:DIR
   # DIR is the source directory for all the snapshots in this repository.
   # HOST is the host where the directory lives.
   if (! open SOURCE, "$repository/$source_info_file") {
      diag "couldn't open $repository/$source_info_file: $!";
      $errors++;
      next;
   }
   my $source = <SOURCE>;
   close SOURCE;
   # integrity check the data
   if ($source eq '') {
      diag "nothing found in $repository/$source_info_file";
      $errors++;
      next;
   }
   if ($source !~ /:/) {
      diag "colon missing in source information in $repository/$source_info_file";
      $errors++;
      next;
   }
   chomp $source;
   my $user;
   my ($host, $dir) = split /:/, $source, 2;
   if ($dir eq '') {
      diag "directory information missing in $repository/$source_info_file";
      $errors++;
      next;
   }
   if ($host eq '') {
      diag "host information missing in $repository/$source_info_file";
      $errors++;
      next;
   }
   my $foo;
   ($foo, $host) = split /\@/, $host, 2;
   if ($host eq '') {
      $host = $foo;
   } else {
      $user = $foo
   }
   $host = lc($host);
   # strip off domain if it matches the local host's
   $host =~ s/\.${domain_regexp}$//;
   # strip off leading slashes
   $dir =~ s,^/,,g;
   # convert remaining slashes to dashes
   $dir =~ s,/,-,g;
   $dir = '-ROOT-' if $dir eq '';
   $foo = "$viewdir/$host";
   # create directories, if they don't exist already
   if (! -d $foo && ! mkdir $foo, 0777) {
      fatal "mkdir of $foo failed: $!";
   }
   $foo = "$viewdir/$host/$dir";
   if (! -d $foo && ! mkdir $foo, 0777) {
      fatal "mkdir of $foo failed: $!";
   }
   # read the list of snapshots that are in the repository
   if (! opendir REPOSITORY, $repository) {
      diag "couldn't opendir $repository: $!";
      $errors++;
      next;
   }
   # snapshot directories are the only names in the repository
   # directory that are all digits.
   my @snapshots = grep /^[0-9]+$/, readdir REPOSITORY;
   closedir REPOSITORY;
   # for each snapshot directory create a symlink to it
   my $snap;
   foreach $snap (@snapshots) {
      my ($sec, $min, $hour, $mday, $mon, $year, undef) = localtime($snap);
      $year += 1900;
      $mon += 1;
      # ISO 8601 date format
      my $link = sprintf "%d-%02d-%02d-%02d:%02d:%02d",
		         $year, $mon, $mday, $hour, $min, $sec;
      my $old = "$repository/$snap";
      # If the repository was specified as a relative path, make
      # it non-relative by prepending the current directory.
      # Otherwise the symlink will be broken.
      $old = "$cwd/$old" if $old !~ m,^/,;
      my $new = "$viewdir/$host/$dir/$link";
      # If the link already exists, leave it.
      if (! -l $new && ! symlink $old, $new) {
         diag "symlink of $old to $new failed: $!";
         $errors++;
         next REPOSITORY;
      }
   }
}

# Now look for symlinks in $viewdir that point to directories
# that are no longer in the filesystem.  Any such links must be
# pointing at directories that have been removed by snap-prune or
# snap-baseline.

# Single quotes are around $viewdir to avoid problems if it
# contains spaces.
if (! open LINKS,
           "$find_program '$viewdir' -type l -name '*-??-??-??:??:??' |") {
   fatal "couldn't generate symlink list with $find_program: $!";
}

my $link;
while (defined ($link = <LINKS>)) {
   chomp $link;
   unlink $link if ! -d $link;
}

close LINKS;

# Look for empty directories in $viewdir and remove them.
# Using find's -d option is crucial here so that we remove directories
# at the bottom of the tree before attempting to remove their
# parents.
if ($find_properties{accepts_dash_d}) {
   if (! open DIRS, "$find_program -d '$viewdir' -type d |") {
      fatal "couldn't generate directory list with $find_program: $!";
   }
} else {
   if (! open DIRS, "$find_program '$viewdir' -depth -type d |") {
      fatal "couldn't generate directory list with $find_program: $!";
   }
}

my $dir;
while (defined ($dir = <DIRS>)) {
   chomp $dir;
   rmdir $dir; # will fail if directory isn't empty
}

close DIRS;

exit 1 if $errors;
exit 0;
